/*
 * Copyright (c) 2020 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

// Package k8s encapsulates interaction with the Kubernetes API. It
// implements the client-go watcher that obtains information about
// TLS Secrets, and updates any PEM files maintained on the current
// system.
package k8s

import (
	"context"
	"fmt"
	"time"

	"code.uplex.de/k8s/k8s-crt-dnldr/pkg/crt"
	"code.uplex.de/k8s/k8s-crt-dnldr/pkg/pem"
	"github.com/sirupsen/logrus"

	"k8s.io/apimachinery/pkg/api/meta"
	"k8s.io/apimachinery/pkg/fields"
	"k8s.io/apimachinery/pkg/runtime"
	utilruntime "k8s.io/apimachinery/pkg/util/runtime"
	"k8s.io/apimachinery/pkg/util/wait"
	"k8s.io/client-go/informers"
	"k8s.io/client-go/kubernetes"
	core_v1 "k8s.io/client-go/kubernetes/typed/core/v1"
	"k8s.io/client-go/tools/cache"
	"k8s.io/client-go/tools/record"

	api_v1 "k8s.io/api/core/v1"
	meta_v1 "k8s.io/apimachinery/pkg/apis/meta/v1"
)

// SyncType classifies the sync event, passed through to workers.
type SyncType uint8

const (
	// Add event
	Add SyncType = iota
	// Update event
	Update
	// Delete event
	Delete
)

func (t SyncType) String() string {
	switch t {
	case Add:
		return "add"
	case Update:
		return "update"
	case Delete:
		return "delete"
	default:
		return "unknown sync type"
	}
}

// SyncObj wraps the Secret for which event handlers are notified, and
// encodes the sync event. These are the objects passed into the
// queues for workers.
type SyncObj struct {
	Type   SyncType
	Secret *api_v1.Secret
}

// Filters Secrets with the type field that identify a Secret to be
// used for TLS certificates, and named in an Ingress resource.
func ingressTLSSecrets(opts *meta_v1.ListOptions) {
	opts.FieldSelector = fields.OneTermEqualSelector("type",
		string(api_v1.SecretTypeTLS)).String()
}

// Watcher watches Kubernetes API for TLS Secrets, and updates or
// deletes PEM files when needed.
type Watcher struct {
	log      *logrus.Logger
	informer cache.SharedIndexInformer
	nsQs     *NamespaceQueues
	ctx      context.Context
	cancel   context.CancelFunc
	getter   *crt.Getter
	files    *pem.Files
}

// NewWatcher creates an API client.
//
//    log: logger initialized at startup
//    kubeClient: k8s client initialized at startup
//    ns: namespace in which to watch Secrets (empty for all)
//    base: directory into which PEM files are stored
//    gid: group ID to be set in PEM file permissions, if >= 0
//    resyncPeriod: for the k8s informer (0s for no resync)
func NewWatcher(
	log *logrus.Logger,
	kubeClient kubernetes.Interface,
	ns string,
	base string,
	gid int,
	resyncPeriod time.Duration,
) (*Watcher, error) {
	watcher := Watcher{
		log: log,
	}

	infFactory := informers.NewSharedInformerFactoryWithOptions(
		kubeClient, resyncPeriod, informers.WithNamespace(ns),
		informers.WithTweakListOptions(ingressTLSSecrets))
	watcher.informer = infFactory.Core().V1().Secrets().Informer()
	lister := infFactory.Core().V1().Secrets().Lister()

	eventBroadcaster := record.NewBroadcaster()
	eventBroadcaster.StartLogging(watcher.log.Printf)
	eventBroadcaster.StartRecordingToSink(&core_v1.EventSinkImpl{
		Interface: kubeClient.CoreV1().Events(ns),
	})
	evtScheme := runtime.NewScheme()
	if err := api_v1.AddToScheme(evtScheme); err != nil {
		return nil, err
	}
	recorder := eventBroadcaster.NewRecorder(evtScheme,
		api_v1.EventSource{Component: "crt-dnldr"})

	evtFuncs := cache.ResourceEventHandlerFuncs{
		AddFunc:    watcher.addObj,
		DeleteFunc: watcher.deleteObj,
		UpdateFunc: watcher.updateObj,
	}

	watcher.informer.AddEventHandler(evtFuncs)

	watcher.getter = crt.NewGetter(lister)
	var err error
	watcher.files, err = pem.NewFiles(base, gid, watcher.getter)
	if err != nil {
		return nil, err
	}
	watcher.nsQs = NewNamespaceQueues(watcher.log, watcher.files, recorder)

	// InitMetrics()

	return &watcher, nil
}

// GetCrtGetter returns a crt.Getter object from the SecretLister
// created by the Watcher. This object should be used exclusively to
// access Secret contents.
func (watcher *Watcher) GetCrtGetter() *crt.Getter {
	return watcher.getter
}

// GetFiles returns a pem.Files that manages PEM files and their metadata.
func (watcher *Watcher) GetFiles() *pem.Files {
	return watcher.files
}

func (watcher *Watcher) getSecret(obj interface{}) (*api_v1.Secret, bool) {
	secret, ok := obj.(*api_v1.Secret)
	if !ok {
		// XXX error counter
		m, mErr := meta.Accessor(obj)
		t, tErr := meta.TypeAccessor(obj)
		if mErr == nil && tErr == nil {
			if t.GetKind() != "" {
				watcher.log.Errorf("Not a Secret: %s %s/%s",
					t.GetKind(), m.GetNamespace(),
					m.GetName())
			} else {
				watcher.log.Errorf("Not a Secret: %s/%s",
					m.GetNamespace(), m.GetName())
			}
		} else {
			watcher.log.Errorf("Got a non-Secret, unknown type")
		}
		return nil, false
	}
	if secret.Type != api_v1.SecretTypeTLS {
		watcher.log.Errorf("Not a TLS Secret: %s/%s", secret.Namespace,
			secret.Name)
		return nil, false
	}
	return secret, true
}

func (watcher *Watcher) logObj(action string, secret *api_v1.Secret) {
	watcher.log.Debugf("%s Secret: %s/%s %s %s", action, secret.Namespace,
		secret.Name, string(secret.UID), secret.ResourceVersion)
}

func (watcher *Watcher) addObj(obj interface{}) {
	secret, ok := watcher.getSecret(obj)
	if !ok {
		return
	}
	watcher.logObj("Add", secret)
	// watchCounters.WithLabelValues("Add").Inc()
	// watcher.nsQs.Queue.Add(&SyncObj{Type: Add, Secret: secret})
}

func (watcher *Watcher) deleteObj(obj interface{}) {
	secret, ok := watcher.getSecret(obj)
	if !ok {
		return
	}
	watcher.logObj("Delete", secret)
	// watchCounters.WithLabelValues("Delete").Inc()
	watcher.nsQs.Queue.Add(&SyncObj{Type: Delete, Secret: secret})
}

func (watcher *Watcher) updateObj(old, new interface{}) {
	newSecret, ok := watcher.getSecret(new)
	if !ok {
		return
	}
	watcher.logObj("Update:", newSecret)
	// watchCounters.WithLabelValues("Update").Inc()
	watcher.nsQs.Queue.Add(&SyncObj{Type: Update, Secret: newSecret})
}

// Run the API client -- start the informers in goroutines, wait for
// the caches to sync, and call Run() for the NamespaceQueues. Then
// block until Stop() is invoked.
func (watcher *Watcher) Run(metricsPort uint16) {
	defer utilruntime.HandleCrash()

	watcher.ctx, watcher.cancel = context.WithCancel(context.Background())

	watcher.log.Info("Launching the Secret informer")
	go watcher.informer.Run(watcher.ctx.Done())

	// watcher.log.Infof("Starting metrics listener at port %d", metricsPort)
	// go ServeMetrics(watcher.log, metricsPort)

	watcher.log.Info("Waiting for cache to sync")
	if ok := cache.WaitForCacheSync(watcher.ctx.Done(),
		watcher.informer.HasSynced); !ok {

		err := fmt.Errorf("Failed waiting for caches to sync")
		utilruntime.HandleError(err)
		return
	}

	watcher.log.Info("Cache synced, running workers")
	// XXX how to get initial state?
	// XXX start http listener
	go wait.Until(watcher.nsQs.Run, time.Second, watcher.ctx.Done())
}

// Stop the k8s watcher -- signal the workers to stop.
func (watcher *Watcher) Stop() {
	watcher.log.Info("Stopping k8s API watcher")
	watcher.cancel()
	<-watcher.ctx.Done()
	watcher.log.Info("Shutting down workers")
	watcher.nsQs.Stop()
	watcher.log.Info("k8s API client exiting")
}
